#include <linux/kernel.h>
#include <linux/delay.h>
#include "ag7100.h"
#include "ag7100_phy.h"

#include "ksz_phy.h"

#define ETH_PHY_ADDR            0

#define KSZ_BASIC_CTRL     0
#define KSZ_BASIC_STATUS   1
#define KSZ_PHY_ID1        2
#define KSZ_PHY_ID2        3
#define KSZ_AUTONEG_ADV    4
#define KSZ_AUTONEG_LPA    5
#define KSZ_AUTONET_EXP    6
#define KSZ_AUTONEG_NP     7
#define KSZ_LP_NPA         8
#define KSZ_RXERR_CNT      0x15
#define KSZ_INT_CTL        0x1B
#define KSZ_PHY_CTL1       0x1E
#define KSZ_PHY_CTL2       0x1F

#define KSZ_RESET          (1 << 15)
#define KSZ_LOOP           (1 << 14)
#define KSZ_LSB            (1 << 13)
#define KSZ_AUTONEG_ENA    (1 << 12)
#define KSZ_POWER_DOWN     (1 << 11)
#define KSZ_ISOLATE        (1 << 10)
#define KSZ_AUTONEG_RST    (1 << 9)
#define KSZ_DUPLEX         (1 << 8)
#define KSZ_COL_TEST       (1 << 7)
#define KSZ_DISABLE        (1)

#define PHYCTRL_MODE                      0x001C
#define PHYCTRL_MODE_AUTONEG             (0 << 2)
#define PHYCTRL_MODE_10BASET_HD          (1 << 2)
#define PHYCTRL_MODE_100BASETX_HD        (2 << 2)
#define PHYCTRL_MODE_10BASET_FD          (5 << 2)
#define PHYCTRL_MODE_100BASETX_FD        (6 << 2)
#define PHYCTRL_MODE_PHY_ISOLATE         (7 << 2)
#define PHYCTRL_AUTONEG_COMPLETE         (1 << 7)

#define ST_100FDX_CAP                    (1 << 14)
#define ST_100HDX_CAP                    (1 << 13)
#define ST_10FDX_CAP                     (1 << 12)
#define ST_10HDX_CAP                     (1 << 11)
#define ST_CAN_AUTONEG                   (1 << 3)
#define ST_AUTONEG_DONE                  (1 << 5)
#define ST_LINK                          (1 << 2)

#define AN_100FDX_CAP                    (1 << 8)
#define AN_100HDX_CAP                    (1 << 7)
#define AN_10FDX_CAP                     (1 << 6)
#define AN_10HDX_CAP                     (1 << 5)

#define RESET_RETRIES 16
#define AUTONEG_RETRIES 8

void ksz_phy_setup(int unit) {
  int i;

/*
  for (i=0; i<0x40; i++) {
    printk("i=%d id=%04x %04x\n",i,ag7100_mii_read(unit, i, KSZ_PHY_ID1),ag7100_mii_read(unit, i, KSZ_PHY_ID2));
  }

  return;  
*/  
  i=ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL) | KSZ_RESET ;
  ag7100_mii_write(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL, i);
  
  for (i=0; i<RESET_RETRIES; i++) {
    if ((ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL) & KSZ_RESET)==0) break;
    mdelay(200);
  }
  
  if (ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL) & KSZ_RESET) {
    printk(KERN_ERR "ksz_phy: reset does not completed\n");
    return;
  }
 
  if (ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_STATUS) & ST_CAN_AUTONEG) {
    int autoneg_ok=0;
    int autoneg_st;
    i=ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL) | KSZ_AUTONEG_ENA|KSZ_AUTONEG_RST;
    ag7100_mii_write(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL, KSZ_AUTONEG_ENA|KSZ_AUTONEG_RST);
    for (i=0; i<AUTONEG_RETRIES; i++) {
      if ((ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_CTRL) & KSZ_AUTONEG_RST)==0 &&
          ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_STATUS) & ST_AUTONEG_DONE) { autoneg_ok++; break; }
      mdelay(200);
    }
  
    if (!autoneg_ok) {
      printk(KERN_ERR "ksz_phy: autonegotiation does not completed\n");
      return;
    }
    
    autoneg_st=ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_AUTONEG_LPA);
    printk(KERN_INFO "ksz_phy: Link partner ability [");
    if (autoneg_st & AN_100FDX_CAP) printk("100FDX ");
    if (autoneg_st & AN_100HDX_CAP) printk("100HDX ");
    if (autoneg_st & AN_10FDX_CAP) printk("10FDX ");
    if (autoneg_st & AN_10HDX_CAP) printk("10HDX ");
    printk("]\n");
  } else printk(KERN_INFO "ksz_phy: autonegotiation is not supported\n");

  return;
}

int ksz_phy_is_up(int unit) {
  return (ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_BASIC_STATUS) & ST_LINK)==ST_LINK;
}

int ksz_phy_is_fdx(int unit) {
  return 1;
  switch (ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_PHY_CTL2) & PHYCTRL_MODE) {
    case PHYCTRL_MODE_10BASET_FD:
    case PHYCTRL_MODE_100BASETX_FD:
      return 1;
    default:
      return 0;
  }
}

int ksz_phy_speed(int unit) {
  switch (ag7100_mii_read(unit, ETH_PHY_ADDR, KSZ_PHY_CTL2) & PHYCTRL_MODE) {
    case PHYCTRL_MODE_10BASET_FD:
    case PHYCTRL_MODE_10BASET_HD:
      return AG7100_PHY_SPEED_10T;
    case PHYCTRL_MODE_100BASETX_FD:
    case PHYCTRL_MODE_100BASETX_HD:
      return AG7100_PHY_SPEED_100TX;
    default:
      return 0;
  }
}
